/*******************************************************************************
**	IBM RGB52x ramdac driver
********************************************************************************
**
   Copyright (C)	1997,98		Paul Sargent
**
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**
********************************************************************************
**	This code was derived from the following sources of information:
**
**	[1]	Texas Intruments TVP 3026 RAMDAC driver for KGI 
**		(C) 1997 Jonas Borgstrom & Steffen Seeger
**	[2]	IBM RGB526/RGB526DB Preliminary Datasheet 95/09/25 
**		(C) 1995 International Business Machines Corp.
*/


/* If you have a ramdac with a higher Mhz then 170 MHz, then use
 * the module parameter 'dacmhz', ex: 'insmod ggidrv.o dacmhz=220'
 */

#include <kgi/maintainers.h>

#define	MAINTAINER	David_Waite
#define	DRIVER_NAME	"RGB52x DAC driver"
#define	DRIVER_REV	"$Revison:$"

#ifndef DEBUG_LEVEL
#define	DEBUG_LEVEL	0	/* do silent checking */
#endif

#ifdef DEBUG
#define	NEED_GRAPHTYPE_SYMS
#endif

#include <kgi/module.h>
#include "IBM/rgb52x.h"

#define DAC_PRECISION	8	/* do 8 bit R,G,B resolution */
#define	SUPPORTED (GM_TEXT16 | GM_4BIT | GM_8BIT | GM_15BIT | GM_16BIT | GM_24BIT | GM_32BIT)


/*
** Module Parameters- max is up to 220 mhz just because the clock is responsible for lowering.
*/

static struct kgi_ramdac rgb52x_info =
{
	"IBM",				/* manufacturer		*/
	"RGB52x",			/* model		*/
	SUPPORTED,			/* supported modes	*/
	FL_DAC_8BIT | FL_DAC_GAMMA,	/* flags		*/
	{ 0x0FFF, 0x0FFF },		/* maximum resolution	*/
	NULL,				/* Private data hook	*/
	DAC_32BIT_PB | DAC_64BIT_PB |	/* device type		*/
	DAC_VRAM,
	 {        0, 220000000 },	/* normal mode		*/
	 {        0,         0 },	/* 2xclk mode		*/

	{{	  0, 220000000 },	/* 	KGIGT_TEXT		*/
	 {        0,         0 },	/*	KGIGT_TEXT32	*/
	 {        0, 220000000 },	/* 	KGIGT_1BIT		*/
	 {	  0, 220000000 },	/* 	KGIGT_4BIT		*/
	 {        0, 220000000 },	/* 	KGIGT_8BIT		*/	
	 {	  0, 220000000 },	/* 	KGIGT_15BIT	*/
	 { 	  0, 220000000 },	/* 	KGIGT_16BIT	*/
	 {    	  0, 220000000 },	/* 	KGIGT_24BIT	*/
 	 {	  0, 220000000 }}	/* 	KGIGT_32BIT	*/
};

/*
**	driver state info
*/

struct rgb52x_private 
{
	kgi_uint	cur_hotx, cur_hoty;	/* cursor hot spot	*/
	kgi_uint	cur_x, cur_y;		/* current position	*/
	kgi_uint	type;				/*ramdac type*/
};

#define	CUR_HOTX	private->cur_hotx
#define	CUR_HOTY	private->cur_hoty
#define	CUR_X		private->cur_x
#define	CUR_Y		private->cur_y
#define TYPE		private->type


/*
**	helper functions
*/

#define	LCLK(m,d)	((mode->lclk.mul == m) && (mode->lclk.div == d))
#define	GET_PRIVATE	struct rgb52x_private *private = KGIM_PRIVATE(dpy,ramdac)
#include "IBM/clut.inc"		/* use dac_set_clut() from normal VGA */

/*
**	dac_set_pointer- set the pointer to a certain size/shape
*/
static void dac_set_pointer_shape(struct kgi_display *dpy, 
	struct kgi_pointer64x64 *ptr){
 	GET_PRIVATE;
 	kgi_uint i, j, tmp;
 	kgiu8 Curs_Ctrl=0;

	DEBUG3("dac_set_pointer_shape Entered");
   
	/*	update private state
 	*/

   	Curs_Ctrl |= 	_RGB52X_EXT30_CURS_ORDER_L_TO_R|
			_RGB52X_EXT30_CURSOR_SIZE_64;

 	switch (ptr->mode) {

 	case pmWindows:
 		Curs_Ctrl |= _RGB52X_EXT30_CURSOR_MODE_XGA;
 		break;

 	case pmX11:
 		Curs_Ctrl |= _RGB52X_EXT30_CURSOR_MODE_X11;
 		break;

 	case pmThreeColor:
 		Curs_Ctrl |= _RGB52X_EXT30_CURSOR_MODE_3COLOR;
 		break;

 	default:
 		INTERNAL_ERROR;
 	}
 	CUR_HOTX = ptr->hotx;
 	CUR_HOTY = ptr->hoty;

 	EXT_out8(dpy, Curs_Ctrl, DAC_CURS_CTRL);

	EXT_out8(dpy,0,DAC_CURS_HOTSPOT_X);
	EXT_out8(dpy,0,DAC_CURS_HOTSPOT_Y);

 	DEBUG1("set_pointer_shape: hot spot (%i,%i), DAC_CURS_CTRL %i", \
 		CUR_HOTX, CUR_HOTY, Curs_Ctrl);

	/*	set pattern
	*/
	
	/* go into an autoincrementing mode since we are dumping data*/
	EXT_AutoIncrement_Enable(dpy);
	EXT_Address(dpy,DAC_CURS_ARRAY);
	
	for (i = 0;i < 512; i++) {

		tmp = 0;
	        for (j = 0;j < 8; j++) {

		  tmp |= (ptr->and_mask[i] & (1 << j)) << (j+1);
		  tmp |= (ptr->xor_mask[i] & (1 << j)) << j;
		}

		DAC_out8(dpy, (tmp >> 8) & 0xff, DAC_EXT_DATA);
		DAC_out8(dpy, (tmp >> 0) & 0xff, DAC_EXT_DATA);
	}
   
	/*
	**	set colors
	*/

	EXT_Address(dpy,DAC_CURS_COLSTART);
	
	for (i = 0; i < 3; i++) {

		DAC_out8(dpy, ptr->col[i].r >> SHIFT, DAC_EXT_DATA);
		DAC_out8(dpy, ptr->col[i].g >> SHIFT, DAC_EXT_DATA);
		DAC_out8(dpy, ptr->col[i].b >> SHIFT, DAC_EXT_DATA);
	}
	/*back into normal mode*/
	EXT_AutoIncrement_Disable(dpy);
}

/*
**	dac_show_pointer - change pointer to a specified location
*/

static void dac_show_pointer(struct kgi_display *dpy, 
		kgi_uint x, kgi_uint y)
{
	GET_PRIVATE;

	DEBUG3("dac_show_pointer Entered");
   
	if ((CUR_X != x) || (CUR_Y != y)) {

		CUR_X = x;
		CUR_Y = y;
		
		EXT_out8(dpy, x,      DAC_CURS_XLOW);
		EXT_out8(dpy, x >> 8, DAC_CURS_XHIGH);
		EXT_out8(dpy, y,      DAC_CURS_YLOW);
		EXT_out8(dpy, y >> 8, DAC_CURS_YHIGH);
	}
}

/*
**	dac_hide_pointer - turn off pointer drawing
*/

static void dac_hide_pointer(struct kgi_display *dpy)
{
	GET_PRIVATE;

	DEBUG3("dac_hide_pointer Entered");
   
	CUR_X = 0x0FFF;
	CUR_Y = 0x0FFF;

	EXT_out8(dpy, 0xFF, DAC_CURS_XLOW);
	EXT_out8(dpy, 0x0F, DAC_CURS_XHIGH);
	EXT_out8(dpy, 0xFF, DAC_CURS_YLOW);
	EXT_out8(dpy, 0x0F, DAC_CURS_YHIGH);
};


/*
**	KGI interface
*/

int kgim_ramdac_init(struct kgi_card *card, struct kgi_display *dpy)
{
	kgiu8 id, rev;
	int type;
	struct rgb52x_private *private;
        private = kgi_malloc_small(sizeof(*private));
        if (private == NULL) {                
                NOTICE2("failed to allocate private data");
                return -ENOMEM;
        }
	
   
	DEBUG3("kgim_ramdac_init Entered");
   
	DEBUG0(DRIVER_NAME " rev " DRIVER_REV);

	id      = EXT_in8(dpy, DAC_ID);
	rev 	= EXT_in8(dpy, DAC_REV);
	type=-1; /*start with error value*/

	if ((id ==_RGB526_EXT01_PROD_ID)&&(rev==_RGB526_EXT00_REV_LEVEL)) {

		DEBUG0("Detected RGB526 RAMDAC. ID: 0x%2x REV: 0x%2x",id, rev);
		type=RGB526;
	} 
	if ((id ==_RGB524_EXT01_PROD_ID)&&(rev==_RGB524_EXT00_REV_LEVEL)) {

		DEBUG0("Detected RGB524 RAMDAC. ID: 0x%2x REV: 0x%2x",id, rev);
		type=RGB524;
	}
	if ((id ==_RGB524_EXT01_PROD_ID)&&(rev==_RGB524_EXT00_REV_BAD)) {

		DEBUG0("Detected RGB524 RAMDAC. ID: 0x%2x REV: 0x%2x",id, rev);
		DEBUG0("Detected Invert SCLK bug");
		type=RGB524;
	}	
	if (type==-1){
		DEBUG0("unknown chip ID (%.2x).", id);
		return -E(RAMDAC, UNKNOWN);
	}

	if (!(private = kgi_malloc_small(sizeof(*private)))) {

		ERROR("failed to allocate private data!");
		return -ENOMEM;
	}
   
	KGIM_PRIVATE(dpy,ramdac) = private;
	memset(private, 0, sizeof(*private));
	TYPE=type;
   
	card->ramdac = &rgb52x_info;
   
	DEBUG1("  .. %s %s ramdac loaded.", card->ramdac->manufact, \
		card->ramdac->model);
	return EOK;
}

void kgim_ramdac_done(struct kgi_display *dpy)
{
	DEBUG3("kgim_ramdac_done Entered");
   
	kgi_free_small(KGIM_PRIVATE(dpy,ramdac));
	KGIM_PRIVATE(dpy,ramdac) = NULL;
	
	/*look into what is needed to restore everything to vga mode*/
	/*probably a register save/restore*/

	DEBUG0(DRIVER_NAME " removed.");
}

int kgim_ramdac_check_mode (struct kgi_display *dpy, struct kgi_mode *mode, 
			enum kgi_tm_cmd *cmd)
{
	int div,mul;
	int newclock = mode->dclk;
	struct kgi_ramdac *dac = dpy->card->ramdac;

	DEBUG3("kgim_ramdac_check_mode Entered");
   	div=mul=1;
	switch (*cmd) {

	case CMD_PROPOSE:

		DEBUG3("Proposing Mode");

	   	switch(mode->request.graphtype) {
		case KGIGT_TEXT16:	div =  1;	
						break;
		case KGIGT_4BIT:	div = 16;	break;
		case KGIGT_8BIT:	div =  8;	break;
 
		case KGIGT_15BIT:
		case KGIGT_16BIT:	div =  4;	break;
  
		case KGIGT_32BIT:	div =  2;	break;
  
 		case KGIGT_24BIT:	div =  8;	
						mul =  3;
			break;

		default:
			DEBUG2("could not handle mode %x",
				mode->request.graphtype);
			return -E(RAMDAC, UNKNOWN);
		}

		mode->clk.mul = mode->clk.div = 1;
		mode->lclk.div = div;
		mode->lclk.mul = mul;
		DEBUG2("%i:%i multiplex mode", mode->lclk.div, mode->lclk.mul);

		mode->flags &= ~FL_DAC_2XCLOCK;
		mode->flags |= FL_DAC_8BIT;
		if ((mode->request.graphtype == KGIGT_15BIT) || 
			(mode->request.graphtype == KGIGT_16BIT) ||
			(mode->request.graphtype == KGIGT_24BIT) ||
			(mode->request.graphtype == KGIGT_32BIT)) {

			mode->flags |= FL_DAC_GAMMA;
		}
		DEBUG3("Proposal Done");
 		return EOK;

	case CMD_LOWER:

		DEBUG3("Lowering Clock");

	   	if ((newclock < dac->clock[mode->request.graphtype].min) ||
			(newclock < dac->normal.min)) {

			printk("%i Hz DCLK is too low\n", newclock);
			return -E(RAMDAC, UNKNOWN);
		}

		if (dac->clock[mode->request.graphtype].max < newclock) {

    			newclock = dac->clock[mode->request.graphtype].max;
		}

		DEBUG2("dclk = %i", newclock);
		mode->dclk   = newclock;
		DEBUG3("Lowering Clock Done");
		return EOK;

	case CMD_RAISE:

		DEBUG3("Raising Clock");
	   
		if ((newclock > dac->clock[mode->request.graphtype].max) ||
			(newclock > dac->normal.max)) {

			printk("%i Hz DCLK is too high\n", newclock);
			return -E(RAMDAC, UNKNOWN);
		}

		if (newclock < dac->clock[mode->request.graphtype].min) {

			newclock = dac->clock[mode->request.graphtype].min;
		}

		if (newclock < dac->normal.min) {

			newclock = dac->normal.min;
		}

		mode->dclk = newclock;
		DEBUG3("Raising Clock Done");
		return EOK;

     case CMD_CHECK:

		DEBUG3("Checking Mode");
	   
		switch(mode->request.graphtype) {

		case KGIGT_TEXT16:
			if (! LCLK(1,1)) {

				return -E(RAMDAC, NOSUP);
			}
			break;

		case KGIGT_4BIT:
			if (! (LCLK(1,16) || LCLK(1,8))) {

				return -E(RAMDAC, NOSUP);
			}
			break;

		case KGIGT_8BIT:
			if (! (LCLK(1,8) || LCLK(1,4) || LCLK(1,2) ||
				LCLK(1,1))) {
				return -E(RAMDAC, NOSUP);
			}
			break;
 
		case KGIGT_15BIT:
		case KGIGT_16BIT:
			if (! (LCLK(1,4) || LCLK(1,2) )) {
				return -E(RAMDAC, NOSUP);
			}
			break;
  
		case KGIGT_32BIT:
			if (! (LCLK(1,2) || LCLK(1,1))) {
				return -E(RAMDAC, NOSUP);
			}
			break;
  
 		case KGIGT_24BIT:
 			if (! (LCLK(3,8) )) {

 				return -E(RAMDAC, NOSUP);
 			}
 			break;

		default:
			/*unknown or unsupported graphics mode*/
			DEBUG2("could not handle mode %x",
				mode->request.graphtype);
			return -E(RAMDAC, UNKNOWN);
		}

		return EOK;

	default:
		/*unknown command*/
		INTERNAL_ERROR;
		return -E(RAMDAC, UNKNOWN);
	}
}

void kgim_ramdac_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	GET_PRIVATE;
	
        kgiu8	Misc_Ctrl1 = 0,
			Misc_Ctrl2 = 0,
			Misc_Ctrl3 = 0,
			DAC_Opt = 0,
        		Pix_Fmt = 0,
			Ctrl8bpp = 0,
			Ctrl16bpp = 0, 
			Ctrl24bpp = 0,
        		Ctrl32bpp = 0,
			Key_Ctrl = 0;/*526 only*/

	enum kgi_sync synctype = dpy->card->monitor->synctype;

	DEBUG3("kgim_ramdac_set_mode Entered");
   
   	switch(mode->request.graphtype) {

 	case KGIGT_TEXT16:
		Misc_Ctrl2 = 	_RGB52X_EXT71_PORT_SEL_VGA |
				_RGB52X_EXT71_PCLK_SEL_LCLK |
				_RGB52X_EXT71_COL_RES_8BIT;
 		break;

 	case KGIGT_4BIT:
		Misc_Ctrl2 = 	_RGB52X_EXT71_PORT_SEL_VRAM |
				_RGB52X_EXT71_PCLK_SEL_PLL |
				_RGB52X_EXT71_COL_RES_8BIT;
		Pix_Fmt = 	_RGB52X_EXT0A_PIXEL_FORMAT_4BPP;

		switch (mode->lclk.div) {

		case  8:
			Misc_Ctrl1  |= _RGB52X_EXT70_VRAM_SIZE_32;
			break;

		case  16:
			Misc_Ctrl1  |= _RGB52X_EXT70_VRAM_SIZE_64;
			break;

		default:
			INTERNAL_ERROR;
			return;
		}
 		break;

 	case KGIGT_8BIT:
		Misc_Ctrl2 = 	_RGB52X_EXT71_PORT_SEL_VRAM |
				_RGB52X_EXT71_PCLK_SEL_PLL |
				_RGB52X_EXT71_COL_RES_8BIT;
		Pix_Fmt = 	_RGB52X_EXT0A_PIXEL_FORMAT_8BPP;
		Ctrl8bpp =	_RGB52X_EXT0B_B8_DCOL_INDIRECT;
	   
		switch (mode->lclk.div) {

		case  4:
			Misc_Ctrl1  |= _RGB52X_EXT70_VRAM_SIZE_32;
			break;

		case  8:
			Misc_Ctrl1  |= _RGB52X_EXT70_VRAM_SIZE_64;
			break;

		default:
			INTERNAL_ERROR;
			return;
		}
 		break;
	 
	 case KGIGT_16BIT:
	 case KGIGT_15BIT:
		Misc_Ctrl2 = 	_RGB52X_EXT71_PORT_SEL_VRAM |
				_RGB52X_EXT71_PCLK_SEL_PLL |
				_RGB52X_EXT71_COL_RES_8BIT;
		Pix_Fmt = 	_RGB52X_EXT0A_PIXEL_FORMAT_16BPP;
	   	Ctrl16bpp = 	(mode->request.graphtype == KGIGT_16BIT) ? 
				_RGB52X_EXT0C_B16_565 : _RGB52X_EXT0C_B16_555;
	   	Ctrl16bpp |= 0xc0;
			/*Direct mode, since the palette isn't 
			getting initialized*/
		/*may also want to change to linear color rather than zero filled*/

		Key_Ctrl = _RGB526_EXT78_KEY_DISABLE;
	   
		switch (mode->lclk.div) {

		case  2:
			Misc_Ctrl1  |= _RGB52X_EXT70_VRAM_SIZE_32;
			break;

		case  4:
			Misc_Ctrl1  |= _RGB52X_EXT70_VRAM_SIZE_64;
			break;

		default:
			INTERNAL_ERROR;
			return;
		}
 		break;
  
 	case KGIGT_24BIT:
		Misc_Ctrl2 = 	_RGB52X_EXT71_PORT_SEL_VRAM |
				_RGB52X_EXT71_PCLK_SEL_PLL |
				_RGB52X_EXT71_COL_RES_8BIT;
		Pix_Fmt = 	_RGB52X_EXT0A_PIXEL_FORMAT_24BPP;
		Ctrl24bpp = 	_RGB52X_EXT0D_B24_DCOL_DIRECT;
	   
		switch (mode->lclk.div) {

		case  8:
			Misc_Ctrl1  |= _RGB52X_EXT70_VRAM_SIZE_64;
			break;
		default:
			INTERNAL_ERROR;
			return;
		}
 		break;

 	case KGIGT_32BIT:
		Misc_Ctrl2 =	_RGB52X_EXT71_PORT_SEL_VRAM |
				_RGB52X_EXT71_PCLK_SEL_PLL |
				_RGB52X_EXT71_COL_RES_8BIT;
		Pix_Fmt = 	_RGB52X_EXT0A_PIXEL_FORMAT_32BPP;
		Ctrl32bpp = 	_RGB52X_EXT0E_B32_DCOL_INDIRECT;
		Key_Ctrl = 	_RGB526_EXT78_KEY_DISABLE;
	   
		switch (mode->lclk.div) {

		case  1:
			Misc_Ctrl1  |= _RGB52X_EXT70_VRAM_SIZE_32;
			break;

		case  2:
			Misc_Ctrl1  |= _RGB52X_EXT70_VRAM_SIZE_64;
			break;

		default:
			INTERNAL_ERROR;
			return;
		}
 		break;

 	default:
 		INTERNAL_ERROR;
 		DEBUG2("could not handle mode %x", mode->request.graphtype);
 		return;
 	}	
	
	EXT_out8(dpy, Misc_Ctrl1 , DAC_MISC_CTRL_1);
	EXT_out8(dpy, Misc_Ctrl2, DAC_MISC_CTRL_2);
	EXT_out8(dpy, Misc_Ctrl3, DAC_MISC_CTRL_3);
   	EXT_out8(dpy, Pix_Fmt, DAC_PIX_FMT);
   	EXT_out8(dpy, Ctrl8bpp, DAC_8BPP_CTRL);
   	EXT_out8(dpy, Ctrl16bpp, DAC_16BPP_CTRL);
   	EXT_out8(dpy, Ctrl24bpp, DAC_24BPP_CTRL);
   	EXT_out8(dpy, Ctrl32bpp, DAC_32BPP_CTRL);
	EXT_out8(dpy, _RGB52X_EXT02_PLL_ENABLE|
		_RGB52X_EXT02_DDOT_PLL_DIV_2, DAC_MISC_CLOCK);

	if ((TYPE==0)||(TYPE==1)) /*only if 526*/
   	EXT_out8(dpy, Key_Ctrl, DAC_KEY_CTRL);

 	EXT_out8(dpy, _RGB52X_EXT30_CURSOR_MODE_OFF, DAC_CURS_CTRL);
	DEBUG3("final- %x %x %x %x %x %x %x %x",
	EXT_in8(dpy, DAC_MISC_CTRL_1),
	EXT_in8(dpy, DAC_MISC_CTRL_2),
	EXT_in8(dpy, DAC_MISC_CTRL_3),
   	EXT_in8(dpy, DAC_PIX_FMT),
   	EXT_in8(dpy, DAC_8BPP_CTRL),
   	EXT_in8(dpy, DAC_16BPP_CTRL),
   	EXT_in8(dpy, DAC_24BPP_CTRL),
   	EXT_in8(dpy, DAC_32BPP_CTRL));

 	if (synctype & (SYNC_ON_RED | SYNC_ON_BLUE)) {
		
 		DEBUG0("Sync on Red or Blue not supported");
		INTERNAL_ERROR;
 		return;
 	}

	DAC_Opt = _RGB52X_EXT06_SLEW_FAST;
   	if (synctype & SYNC_ON_GREEN) {

	 	DAC_Opt |= _RGB52X_EXT06_SOG_ENABLE;
 	}

 	if (synctype & SYNC_PEDESTRAL) {

	 	DAC_Opt |= _RGB52X_EXT06_DPE_ENABLE;
 	}
   
 	EXT_out8(dpy, DAC_Opt, DAC_DAC_OP);

 	EXT_out8(dpy, 0x00, DAC_PAL); 

 	CUR_X = 0xFFF; 
 	CUR_Y = 0xFFF;
 	EXT_out8(dpy, 0xFF, DAC_CURS_XLOW);
 	EXT_out8(dpy, 0x0F, DAC_CURS_XHIGH);
 	EXT_out8(dpy, 0xFF, DAC_CURS_YLOW);
 	EXT_out8(dpy, 0x0F, DAC_CURS_YHIGH);

        dpy->set_clut = dac_set_clut;
 	dpy->set_pointer_shape  = dac_set_pointer_shape;
        dpy->show_pointer = dac_show_pointer;
        dpy->hide_pointer = dac_hide_pointer;
 	dpy->undo_pointer = NULL;
}

int kgim_ramdac_command (struct kgi_graphic *graph, kgi_uint cmd, kgi_uint arg)
{
	/*	this may set 
	**	-  dac_params (take care if 504 or 505 was detected).
	**	-  hardware cursor
	**	-  CLUT (or gamma correction table).
	*/

	kgi_clut clut;

	DEBUG3("kgim_ramdac_command Entered");
   
	switch(cmd) {

	case RAMDAC_SETCLUT:
 		clut.size = 256;
 		clut.data=(kgi_color *)(graph->io_buf);
 		dac_set_clut(graph->dev.dpy,&clut,0,256);
 		return 0;
	case RAMDAC_GETCLUT:
 		clut.size=256;
 		clut.data=(kgi_color *)(graph->io_buf);
 		graph->io_res=graph->io_buf;
 		dac_get_clut(graph->dev.dpy,&clut,0,256);
 		return 0;
	}

	/*	This driver can't do this. Maybe a simpler command will do.
	*/
	return -E(RAMDAC,NOSUP_ALWAYS_LOWER);	
}

/*
**	Register reporting function
*/

void ramdac_report(struct kgi_display *dpy){
	int i;
	printk("IBMRGB register dump\n");
	printk("--: 00 01 02 03 04 05 06 07  08 09 0a 0b 0c 0d 0e 0f\n");
	for (i=0;i<0x100;i++){
		if (i%0x10==0) printk("\n%.2x-",i);
		if (i%0x10==8) printk(" ");
		printk(" %.2x",EXT_in8(dpy,i));
	}
}
